# coding=utf-8
from __future__ import unicode_literals

import datetime
import traceback
import uuid
from multiprocessing.pool import Pool
from unittest.case import TestCase

import six

from eventsourcing.contrib.suffixtrees.application import SuffixTreeApplicationWithPythonObjects, \
    SuffixTreeApplicationWithCassandra, AbstractSuffixTreeApplication
from eventsourcing.contrib.suffixtrees.domain.model.generalizedsuffixtree import GeneralizedSuffixTree, \
    SuffixTreeEdge, SuffixTreeNode, STRING_ID_END
from eventsourcing.tests.unit_test_fixtures_suffix_tree_text import LONG_TEXT, LONG_TEXT_CONT
from eventsourcing.tests.unit_test_cases_cassandra import CassandraTestCase


class GeneralizedSuffixTreeTestCase(TestCase):

    def setUp(self):
        super(GeneralizedSuffixTreeTestCase, self).setUp()
        self.app = SuffixTreeApplicationWithPythonObjects()

    def tearDown(self):
        self.app.close()
        super(GeneralizedSuffixTreeTestCase, self).tearDown()

    def add_string_to_suffix_tree(self, string, string_id, suffix_tree):
        print("Adding string to suffix tree: {}".format(repr(string[:100])))
        started = datetime.datetime.now()
        suffix_tree.add_string(string, string_id=string_id)
        print(" - added string in: {}".format(datetime.datetime.now() - started))



class TestGeneralizedSuffixTreeFast(GeneralizedSuffixTreeTestCase):

    def test_empty_string(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        self.add_string_to_suffix_tree('', '1', st)
        self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
        self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
        self.assertFalse(self.app.has_substring('not there', st.id))
        self.assertFalse(self.app.has_substring('', st.id))

    def test_add_prefix_of_previous(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        self.add_string_to_suffix_tree('aba', '1', st)
        self.add_string_to_suffix_tree('a', '2', st)

        # Check the string ID is returned.
        result_ids = self.app.find_string_ids('a', st.id)
        self.assertEqual(result_ids, {'1', '2'})

    def test_given_add_duplicate_when_query_with_suffix_then_results_have_both(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        self.add_string_to_suffix_tree('ab', '1', st)
        self.add_string_to_suffix_tree('ab', '2', st)

        # Check the string ID is returned.
        result_ids = self.app.find_string_ids('b', st.id)
        self.assertEqual(result_ids, {'1', '2'})

    def test_extended_copy_of_previous_string(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        self.add_string_to_suffix_tree('a', '1', st)
        self.add_string_to_suffix_tree('ab', '2', st)

        # Check the common prefix can be found.
        result_ids = self.app.find_string_ids('a', st.id)
        self.assertEqual(result_ids, {'1', '2'})

        # Check the string ID is returned.
        result_ids = self.app.find_string_ids('ab', st.id)
        self.assertEqual(result_ids, {'2'})

    def test_trucated_copy_of_previous_string(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        self.add_string_to_suffix_tree('1a', '2', st)
        self.add_string_to_suffix_tree('1ab', '1', st)

        # Check the common prefix can be found.
        result_ids = self.app.find_string_ids('a', st.id)
        self.assertEqual(result_ids, {'1', '2'})

        # Check the string ID is returned.
        result_ids = self.app.find_string_ids('ab', st.id)
        self.assertEqual(result_ids, {'1'})

        # Check the string ID is returned.
        result_ids = self.app.find_string_ids('b', st.id)
        self.assertEqual(result_ids, {'1'})

    def test_extended_copy_of_longer_previous_string(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        self.add_string_to_suffix_tree('ab', '1', st)
        self.add_string_to_suffix_tree('abc', '2', st)

        # Check the common prefix can be found.
        result_ids = self.app.find_string_ids('a', st.id)
        self.assertEqual(result_ids, {'1', '2'})

        # Check the common prefix can be found.
        result_ids = self.app.find_string_ids('ab', st.id)
        self.assertEqual(result_ids, {'1', '2'})

        # Check the extention can be found.
        result_ids = self.app.find_string_ids('c', st.id)
        self.assertEqual(result_ids, {'2'})

        # Check the extended string can be found.
        result_ids = self.app.find_string_ids('abc', st.id)
        self.assertEqual(result_ids, {'2'})

    def test_non_repeating_string(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        string_id = '1'
        string = 'abc'
        self.add_string_to_suffix_tree(string, string_id, st)

        # Check various non-substrings aren't in the tree.
        self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
        self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
        self.assertFalse(self.app.has_substring('not there', st.id))
        self.assertFalse(self.app.has_substring('', st.id))

        # Check substrings are in the tree.
        edge, ln = self.app.find_substring_edge('abc', st.id)
        assert isinstance(edge, SuffixTreeEdge)
        # self.assertEqual(edge.label, string + STRING_ID_END)
        # self.assertEqual(edge.label, string + string_id + STRING_ID_END)

        # Check the edge has a node.
        node = self.app.node_repo[edge.dest_node_id]
        assert isinstance(node, SuffixTreeNode)

        # Check the node has a string ID.
        self.assertEqual(node.string_id, string_id)

        # Check the string ID is returned.
        result_ids = self.app.find_string_ids(string, st.id)
        self.assertEqual(result_ids, {string_id})

    def test_repeating_string(self):
        st = self.app.register_new_suffix_tree()
        self.add_string_to_suffix_tree("aaa", '1', st)

        edge, ln = self.app.find_substring_edge('a', st.id)
        # self.assertEqual(edge.label, 'a')
        edge, ln = self.app.find_substring_edge('aa', st.id)
        # self.assertEqual(edge.label, 'a')
        edge, ln = self.app.find_substring_edge('aaa', st.id)
        # self.assertEqual(edge.label, 'a1' + STRING_ID_END)
        edge, ln = self.app.find_substring_edge('b', st.id)
        self.assertEqual(edge, None)
        self.assertTrue(self.app.has_substring('a', st.id))
        self.assertTrue(self.app.has_substring('aa', st.id))
        self.assertTrue(self.app.has_substring('aaa', st.id))
        self.assertFalse(self.app.has_substring('aaaa', st.id))
        self.assertFalse(self.app.has_substring('b', st.id))
        # case sensitive by default
        self.assertFalse(self.app.has_substring('A', st.id))

        # Check ID is returned.
        self.assertEqual(self.app.find_string_ids('a', st.id), {'1'})
        self.assertEqual(self.app.find_string_ids('aa', st.id), {'1'})
        self.assertEqual(self.app.find_string_ids('aaa', st.id), {'1'})

    def test_two_identical_strings(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        string1_id = '1'
        string1 = 'abc'
        string2_id = '2'
        string2 = 'abc'
        self.add_string_to_suffix_tree(string1, string1_id, st)
        self.add_string_to_suffix_tree(string2, string2_id, st)

        # Check various non-substrings aren't in the tree.
        self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
        self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
        self.assertFalse(self.app.has_substring('not there', st.id))
        self.assertFalse(self.app.has_substring('', st.id))

        # Check substrings of string1 are in the tree.
        edge, ln = self.app.find_substring_edge(string1, st.id)
        assert isinstance(edge, SuffixTreeEdge)
        # self.assertEqual(edge.label, string1 + STRING_ID_END)
        # self.assertEqual(edge.label, string1 + string1_id + STRING_ID_END)
        # self.assertEqual(edge.label, string1 + string1_id )

        # Check the edge has a node.
        node = self.app.node_repo[edge.dest_node_id]
        assert isinstance(node, SuffixTreeNode)

        # Check the node doesn't have a string ID.
        # self.assertEqual(node.string_id, None)

        # Check both IDs are returned.
        result_ids = self.app.find_string_ids(string1, st.id)
        self.assertEqual(result_ids, {string1_id, string2_id}, (result_ids))

    def test_two_non_repeating_strings(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        string1_id = '1'
        string1 = 'abc'
        string2_id = '2'
        string2 = 'def'
        self.add_string_to_suffix_tree(string1, string1_id, st)
        self.add_string_to_suffix_tree(string2, string2_id, st)

        # Check various non-substrings aren't in the tree.
        self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
        self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
        self.assertFalse(self.app.has_substring('not there', st.id))
        self.assertFalse(self.app.has_substring('', st.id))

        # Check substrings of string1 are in the tree.
        edge, ln = self.app.find_substring_edge(string1, st.id)
        assert isinstance(edge, SuffixTreeEdge)
        # self.assertEqual(edge.label, string1 + STRING_ID_END)
        # self.assertEqual(edge.label, string1 + string1_id + STRING_ID_END)

        # Check the edge has a node.
        node = self.app.node_repo[edge.dest_node_id]
        assert isinstance(node, SuffixTreeNode)

        # Check the node has a string ID.
        self.assertEqual(node.string_id, string1_id)

        # Check the string1 ID is returned.
        result_ids = self.app.find_string_ids(string1, st.id)
        self.assertEqual(result_ids, {string1_id})


        # Check substrings of string2 are in the tree.
        edge, ln = self.app.find_substring_edge(string2, st.id)
        assert isinstance(edge, SuffixTreeEdge)
        # self.assertEqual(edge.label, string2 + STRING_ID_END)
        # self.assertEqual(edge.label, string2 + string2_id+ STRING_ID_END)

        # Check the edge has a node.
        node = self.app.node_repo[edge.dest_node_id]
        assert isinstance(node, SuffixTreeNode)

        # Check the node has a string ID.
        self.assertEqual(node.string_id, string2_id)

        # Check the string2 ID is returned.
        result_ids = self.app.find_string_ids(string2, st.id)
        self.assertEqual(result_ids, {string2_id})

    def test_common_prefix_two_non_repeating_strings_with(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        common_prefix = 'z'
        string1_id = '1'
        string1 = common_prefix + 'abc'
        string2_id = '2'
        string2 = common_prefix + 'def'
        self.add_string_to_suffix_tree(string1, string1_id, st)
        self.add_string_to_suffix_tree(string2, string2_id, st)

        # Check various non-substrings aren't in the tree.
        self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
        self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
        self.assertFalse(self.app.has_substring('not there', st.id))
        self.assertFalse(self.app.has_substring('', st.id))

        # Check substrings of string1 are in the tree.
        edge, ln = self.app.find_substring_edge(string1, st.id)
        assert isinstance(edge, SuffixTreeEdge)
        # self.assertEqual(edge.label, string1[1:] + STRING_ID_END)
        # self.assertEqual(edge.label, string1[1:] + string1_id + STRING_ID_END)

        # Check the edge has a node.
        node = self.app.node_repo[edge.dest_node_id]
        assert isinstance(node, SuffixTreeNode)

        # Check the node has a string ID.
        self.assertEqual(node.string_id, string1_id)

        # Check the string1 ID is returned.
        result_ids = self.app.find_string_ids(string1, st.id)
        self.assertEqual(result_ids, {string1_id})

        # Check substrings of string2 are in the tree.
        edge, ln = self.app.find_substring_edge(string2, st.id)
        assert isinstance(edge, SuffixTreeEdge)
        # self.assertEqual(edge.label, string2[1:] + STRING_ID_END)
        # self.assertEqual(edge.label, string2[1:] + string2_id + STRING_ID_END)

        # Check the edge has a node.
        node = self.app.node_repo[edge.dest_node_id]
        assert isinstance(node, SuffixTreeNode)

        # Check the node has a string ID.
        self.assertEqual(node.string_id, string2_id)

        # Check the string2 ID is returned.
        result_ids = self.app.find_string_ids(string2, st.id)
        self.assertEqual(result_ids, {string2_id})

        # Check common prefix is in the tree.
        edge, ln = self.app.find_substring_edge(common_prefix, st.id)
        assert isinstance(edge, SuffixTreeEdge)
        self.assertEqual(edge.label, common_prefix)

        # Check the edge has a node.
        node = self.app.node_repo[edge.dest_node_id]
        assert isinstance(node, SuffixTreeNode)

        # Check the node has a string ID.
        self.assertEqual(node.string_id, None)

        # Check both IDs are returned.
        result_ids = self.app.find_string_ids(common_prefix, st.id)
        self.assertEqual(result_ids, {string1_id, string2_id})

    def test_common_suffix_two_non_repeating_strings_with(self):
        st = self.app.register_new_suffix_tree()
        assert isinstance(st, GeneralizedSuffixTree)
        common_suffix = 'z'
        string1_id = '1'
        string1 = 'abc' + common_suffix
        string2_id = '2'
        string2 = 'def' + common_suffix
        self.add_string_to_suffix_tree(string1, string1_id, st)
        self.add_string_to_suffix_tree(string2, string2_id, st)

        # Check various non-substrings aren't in the tree.
        self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
        self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
        self.assertFalse(self.app.has_substring('not there', st.id))
        self.assertFalse(self.app.has_substring('', st.id))

        # Check substrings of string1 are in the tree.
        edge, ln = self.app.find_substring_edge(string1, st.id)
        assert isinstance(edge, SuffixTreeEdge)
        # self.assertEqual(edge.label, string1 + STRING_ID_END)
        # self.assertEqual(edge.label, string1 + string1_id + STRING_ID_END)

        # Check the edge has a node.
        node = self.app.node_repo[edge.dest_node_id]
        assert isinstance(node, SuffixTreeNode)

        # Check the node has a string ID.
        self.assertEqual(node.string_id, string1_id)

        # Check the string1 ID is returned.
        result_ids = self.app.find_string_ids(string1, st.id)
        self.assertEqual(result_ids, {string1_id})

        # Check substrings of string2 are in the tree.
        edge, ln = self.app.find_substring_edge(string2, st.id)
        assert isinstance(edge, SuffixTreeEdge)
        # self.assertEqual(edge.label, string2 + STRING_ID_END)
        # self.assertEqual(edge.label, string2 + string2_id + STRING_ID_END)

        # Check the edge has a node.
        node = self.app.node_repo[edge.dest_node_id]
        assert isinstance(node, SuffixTreeNode)

        # Check the node has a string ID.
        self.assertEqual(node.string_id, string2_id)

        # Check the string2 ID is returned.
        result_ids = self.app.find_string_ids(string2, st.id)
        self.assertEqual(result_ids, {string2_id})

        # Check common suffix is in the tree.
        edge, ln = self.app.find_substring_edge(common_suffix, st.id)
        assert isinstance(edge, SuffixTreeEdge)
        # self.assertEqual(edge.label, common_suffix + STRING_ID_END)
        # self.assertEqual(edge.label, common_suffix)

        # Check the edge has a node.
        node = self.app.node_repo[edge.dest_node_id]
        assert isinstance(node, SuffixTreeNode)

        # Check the node doesn't have a string ID.
        # self.assertEqual(node.string_id, None)

        # Check both IDs are returned.
        result_ids = self.app.find_string_ids(common_suffix, st.id)
        self.assertEqual(result_ids, {string1_id, string2_id})

    def test_mississippi(self):
        st = self.app.register_new_suffix_tree()
        self.add_string_to_suffix_tree("mississippi", "$", st)
        self.add_string_to_suffix_tree("mudpie", "#", st)
        self.add_string_to_suffix_tree("ball", "%", st)
        self.add_string_to_suffix_tree("ball", "&", st)

        # Check there isn't an 'a'.
        self.assertEqual(self.app.find_substring_edge('j', st.id), (None, None))

        # Check 'm'.
        edge, ln = self.app.find_substring_edge('m', st.id)
        # self.assertEqual(edge.label, 'm')
        self.assertEqual(self.app.find_string_ids('m', st.id), {'$', '#'})

        # Check 'mi'.
        edge, ln = self.app.find_substring_edge('mi', st.id)
        # self.assertEqual(edge.label, 'ississippi' + STRING_ID_END)
        self.assertEqual(self.app.find_string_ids('mi', st.id), {'$'})

        # Check 'mu'.
        edge, ln = self.app.find_substring_edge('mu', st.id)
        # self.assertEqual(edge.label, 'udpie' + STRING_ID_END)
        self.assertEqual(self.app.find_string_ids('mu', st.id), {'#'})

        # Check 'missi'.
        edge, ln = self.app.find_substring_edge('missi', st.id)
        self.assertEqual(edge.label, 'ississippi' + STRING_ID_END)

        # Check 'issi'.
        edge, ln = self.app.find_substring_edge('issi', st.id)
        self.assertEqual(edge.label, 'ssi')

        # Check 'is'.
        edge, ln = self.app.find_substring_edge('is', st.id)
        self.assertEqual(edge.label, 'ssi')

        # Check 'si'.
        edge, ln = self.app.find_substring_edge('si', st.id)
        self.assertEqual(edge.label, 'i')

        # Check 'issip'.
        edge, ln = self.app.find_substring_edge('issip', st.id)
        self.assertEqual(edge.label, 'ppi' + STRING_ID_END)

        # Check 'ssip'.
        edge, ln = self.app.find_substring_edge('ssip', st.id)
        self.assertEqual(edge.label, 'ppi' + STRING_ID_END)

        # Check 'sip'.
        edge, ln = self.app.find_substring_edge('sip', st.id)
        self.assertEqual(edge.label, 'ppi' + STRING_ID_END)

        # Check 'ip'.
        edge, ln = self.app.find_substring_edge('ip', st.id)
        self.assertEqual(edge.label, 'ppi' + STRING_ID_END)

        # Check 'i'.
        edge, ln = self.app.find_substring_edge('i', st.id)
        self.assertEqual(edge.label, 'i')

       # Check 'ippi'.
        edge, ln = self.app.find_substring_edge('ippi', st.id)
        self.assertEqual(edge.label, 'ppi' + STRING_ID_END)

        # Check 'mudpie'.
        edge, ln = self.app.find_substring_edge('mudpie', st.id)
        self.assertEqual(edge.label, 'udpie' + STRING_ID_END)

        # Check 'ball'.
        edge, ln = self.app.find_substring_edge('ball', st.id)
        self.assertEqual(edge.label, 'ball' + STRING_ID_END)

        # Check ID is returned.
        self.assertEqual(self.app.find_string_ids('mi', st.id), {'$'})
        self.assertEqual(self.app.find_string_ids('si', st.id), {'$'})
        self.assertEqual(self.app.find_string_ids('pp', st.id), {'$'})
        self.assertEqual(self.app.find_string_ids('mu', st.id), {'#'})
        self.assertEqual(self.app.find_string_ids('m', st.id), {'$','#'})
        self.assertEqual(self.app.find_string_ids('b', st.id), {'%','&'})

    def test_colours(self):
        st = self.app.register_new_suffix_tree()
        self.add_string_to_suffix_tree("blue", "$", st)
        self.add_string_to_suffix_tree("red", "#", st)

        # Check 'b'.
        edge, ln = self.app.find_substring_edge('b', st.id)
        self.assertEqual(edge.label, 'blue' + STRING_ID_END)

        # Check 'l'.
        edge, ln = self.app.find_substring_edge('l', st.id)
        self.assertEqual(edge.label, 'lue' + STRING_ID_END)

        # Check 'u'.
        edge, ln = self.app.find_substring_edge('u', st.id)
        self.assertEqual(edge.label, 'ue' + STRING_ID_END)

        # Check 'e'.
        edge, ln = self.app.find_substring_edge('e', st.id)
        self.assertEqual(edge.label, 'e')

        # Check 'ue'.
        edge, ln = self.app.find_substring_edge('ue', st.id)
        self.assertEqual(edge.label, 'ue' + STRING_ID_END)

        # Check 'lue'.
        edge, ln = self.app.find_substring_edge('lue', st.id)
        self.assertEqual(edge.label, 'lue' + STRING_ID_END)

        # Check 'blue'.
        edge, ln = self.app.find_substring_edge('blue', st.id)
        self.assertEqual(edge.label, 'blue' + STRING_ID_END)

        # Check 're'.
        edge, ln = self.app.find_substring_edge('re', st.id)
        self.assertEqual(edge.label, 'red' + STRING_ID_END)

        # Check 'ed'.
        edge, ln = self.app.find_substring_edge('ed', st.id)
        self.assertEqual(edge.label, 'd' + STRING_ID_END)

        # Check 'red'.
        edge, ln = self.app.find_substring_edge('red', st.id)
        self.assertEqual(edge.label, 'red' + STRING_ID_END)

    def test_find_string_ids(self):
        # This test is the first to involve the children of nodes.
        st = self.app.register_new_suffix_tree()
        string1_id = uuid.uuid4().hex
        string2_id = uuid.uuid4().hex
        string3_id = uuid.uuid4().hex
        # string4_id = uuid.uuid4().hex
        self.add_string_to_suffix_tree("blue", string1_id, st)
        self.add_string_to_suffix_tree("red", string2_id, st)
        self.add_string_to_suffix_tree("blues", string3_id, st)

        strings = self.app.find_string_ids('$', st.id)
        assert not strings, strings
        strings = self.app.find_string_ids('0', st.id)
        assert not strings, strings
        strings = self.app.find_string_ids('1', st.id)
        assert not strings, strings
        strings = self.app.find_string_ids('2', st.id)
        assert not strings, strings
        strings = self.app.find_string_ids('3', st.id)
        assert not strings, strings
        strings = self.app.find_string_ids('4', st.id)
        assert not strings, strings
        strings = self.app.find_string_ids('5', st.id)
        assert not strings, strings
        strings = self.app.find_string_ids('6', st.id)
        assert not strings, strings
        strings = self.app.find_string_ids('7', st.id)
        assert not strings, strings

        # Find 'b'.
        strings = self.app.find_string_ids('b', st.id)
        self.assertIn(string1_id, strings)
        self.assertNotIn(string2_id, strings)

        # Find 'e'.
        strings = self.app.find_string_ids('e', st.id)
        self.assertEqual(len(strings), 3)
        self.assertIn(string1_id, strings, (string1_id, string2_id, string3_id, strings))
        self.assertIn(string2_id, strings, (string1_id, string2_id, string3_id, strings))
        self.assertIn(string3_id, strings, (string1_id, string2_id, string3_id, strings))

        # Find 'e' - limit 1.
        strings = self.app.find_string_ids('e', st.id, limit=1)
        self.assertEqual(len(strings), 1)

        # Find 'e' - limit 2.
        strings = self.app.find_string_ids('e', st.id, limit=2)
        self.assertEqual(len(strings), 2)

        # Find 'r'.
        strings = self.app.find_string_ids('r', st.id)
        self.assertNotIn(string1_id, strings)
        self.assertIn(string2_id, strings)

        # Find 'd'.
        strings = self.app.find_string_ids('d', st.id)
        self.assertNotIn(string1_id, strings)
        self.assertIn(string2_id, strings)

        # Find 's'.
        strings = self.app.find_string_ids('s', st.id)
        self.assertNotIn(string1_id, strings)
        self.assertNotIn(string2_id, strings)
        self.assertIn(string3_id, strings)

    def test_add_string_to_suffixtree_from_repo(self):
        # Check adding strings after getting the tree from the repo.
        st = self.app.register_new_suffix_tree()

        st = self.app.get_suffix_tree(st.id)
        self.add_string_to_suffix_tree('blue', '1', st)

        st = self.app.get_suffix_tree(st.id)
        self.add_string_to_suffix_tree('green', '2', st)

        st = self.app.get_suffix_tree(st.id)
        self.add_string_to_suffix_tree('yellow', '3', st)

        st = self.app.get_suffix_tree(st.id)
        strings_ids = self.app.find_string_ids('e', st.id)
        self.assertEqual(3, len(strings_ids), strings_ids)
        self.assertEqual(['1', '2', '3'], sorted(strings_ids))

    def test_remove_string_from_suffixtree(self):
        st = self.app.register_new_suffix_tree()

        # Add 'blue' and 'green'.
        self.add_string_to_suffix_tree('blue', '1', st)
        self.add_string_to_suffix_tree('green', '2', st)

        strings_ids = self.app.find_string_ids('b', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('l', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('u', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('r', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('g', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('e', st.id)
        self.assertEqual(2, len(strings_ids))

        # Remove 'blue'.
        st.remove_string('blue', '1')

        strings_ids = self.app.find_string_ids('b', st.id)
        self.assertEqual(0, len(strings_ids))

        strings_ids = self.app.find_string_ids('l', st.id)
        self.assertEqual(0, len(strings_ids))

        strings_ids = self.app.find_string_ids('u', st.id)
        self.assertEqual(0, len(strings_ids))

        strings_ids = self.app.find_string_ids('r', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('g', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('e', st.id)
        self.assertEqual(1, len(strings_ids), strings_ids)

        # Add 'blue' again.
        self.add_string_to_suffix_tree('blue', '1', st)

        strings_ids = self.app.find_string_ids('b', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('l', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('u', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('r', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('g', st.id)
        self.assertEqual(1, len(strings_ids))

        strings_ids = self.app.find_string_ids('e', st.id)
        self.assertEqual(2, len(strings_ids))


class TestGeneralizedSuffixTreeSlow(GeneralizedSuffixTreeTestCase):

    def test_long_string(self):
        st = self.app.register_new_suffix_tree()
        self.add_string_to_suffix_tree(LONG_TEXT[:12000], '1', st)
        self.assertEqual(self.app.find_string_ids('Ukkonen', st.id), {'1'})
        self.assertEqual(self.app.find_string_ids('Optimal', st.id), {'1'})
        self.assertFalse(self.app.find_string_ids('ukkonen', st.id))
        self.add_string_to_suffix_tree(LONG_TEXT_CONT[:1000], '2', st)
        self.assertEqual(self.app.find_string_ids('Burrows-Wheeler', st.id), {'2'})
        self.assertEqual(self.app.find_string_ids('suffix', st.id), {'1', '2'})

    def test_split_long_string(self):
        # Split the long string into separate strings, and make some IDs.
        list_of_strings = [w for w in LONG_TEXT[:1000].split(' ') if w]

        print("Long string split into words: {}".format(list_of_strings))

        string_ids = {}
        for string in list_of_strings:
            string_id = uuid.uuid4().hex
            string_ids[string_id] = string

        # Build the suffix tree.
        st = self.app.register_new_suffix_tree()
        for string_id, string in string_ids.items():
            self.add_string_to_suffix_tree(string, string_id, st)

        # Check the suffix tree.
        for string_id, string in string_ids.items():
            self.assertIn(string_id, self.app.find_string_ids(string, st.id), (string, string_id))

        # Build the suffix tree.
        #  - this time, get the suffix tree afresh from the repo each time
        st = self.app.register_new_suffix_tree()
        for string_id, string in string_ids.items():
            st = self.app.get_suffix_tree(st.id)
            self.add_string_to_suffix_tree(string, string_id, st)

        # Check the suffix tree.
        for string_id, string in string_ids.items():
            self.assertIn(string_id, self.app.find_string_ids(string, st.id))

    def test_case_insensitivity(self):
        st = self.app.register_new_suffix_tree(case_insensitive=True)
        self.add_string_to_suffix_tree(LONG_TEXT[:12000], '1', st)
        self.assertEqual(self.app.find_string_ids('ukkonen', st.id), {'1'})
        self.assertEqual(self.app.find_string_ids('Optimal', st.id), {'1'})
        self.assertEqual(self.app.find_string_ids('burrows-wheeler', st.id), set())
        self.add_string_to_suffix_tree(LONG_TEXT_CONT[:1000], '2', st)
        self.assertEqual(self.app.find_string_ids('ukkonen', st.id), {'1'})
        self.assertEqual(self.app.find_string_ids('Optimal', st.id), {'1'})
        self.assertEqual(self.app.find_string_ids('burrows-wheeler', st.id), {'2'})


class TestMultiprocessingWithGeneralizedSuffixTree(CassandraTestCase):

    def setUp(self):
        super(TestMultiprocessingWithGeneralizedSuffixTree, self).setUp()
        self.app = None

    def tearDown(self):
        super(TestMultiprocessingWithGeneralizedSuffixTree, self).tearDown()
        if self.app is not None:
            self.app.close()

    def test_words_in_sorted_order(self):
        self.check_words(is_sorted=True)

    # Todo: Fix this - adding strings in a random order sometimes breaks (perhaps a dict is causing indeterminate order).
    # def test_words_in_unsorted_order(self):
    #     self.check_words()

    # Todo: Fix this - adding strings in a reversed sorted order always fails. Not sure why all substrings of 'ree' fail. The suffix is obviously not moving along in the same way as it does when the nodes are added. Perhaps it needs to add the IDs when explicit match is made, and then move the first char along by one? Not sure so trace it out?
    # def test_words_in_reverse_sorted_order(self):
    #     self.check_words(is_reversed=True)
    #
    # The error reported is:-
    #
    # >       self.assertFalse(errors, "\n".join(errors))
    # E       Not found: substring ''e'' from string ''tree''
    # E       Not found: substring ''ee'' from string ''tree''
    # E       Not found: substring ''r'' from string ''tree''
    # E       Not found: substring ''re'' from string ''tree''
    # E       Not found: substring ''ree'' from string ''tree''

    def check_words(self, is_sorted=False, is_reversed=False):
        # Split the long string into separate strings, and make some IDs.
        words = list([w for w in LONG_TEXT[:100].split(' ') if w])

        print("Adding words: {}".format(words))

        # Avoid adding the same string twice (or a prefix of a previous string).
        #  - because it's a current problem unless we append string IDs, which makes things too slow
        # words = set(words)
        # words = [w for w in words if 0 != sum([x.startswith(w) for x in words if x != w])]

        assert words

        # Make a string ID for each string.
        strings = {}
        for string in words:
            string_id = uuid.uuid4().hex
            strings[string_id] = string

        # Create a new suffix tree.
        self.app = SuffixTreeApplicationWithCassandra()
        st = self.app.register_new_suffix_tree()
        assert st.id in self.app.suffix_tree_repo

        # Close the app, so the pool doesn't inherit it.
        self.app.close()

        # Start the pool.
        pool = Pool(initializer=pool_initializer, processes=1)

        words = [[s, sid, st.id] for sid, s in strings.items() if s]

        if is_sorted:
            words = sorted(words)
        if is_reversed:
            words = reversed(words)

        results = pool.map(add_string_to_suffix_tree, words)
        for result in results:
            if isinstance(result, Exception):
                print(result.args[0][1])
                raise result

        # Creat the app again.
        self.app = SuffixTreeApplicationWithCassandra()

        errors = []

        # Check the suffix tree returns string ID for all substrings of string.
        for string_id, string in strings.items():
            # Check all prefixes and suffixes.
            substrings = sorted(list(get_all_substrings(string)))
            print("")
            print("Checking for all substrings of string '{}': {}".format(repr(string), " ".join([repr(s) for s in substrings])))
            for substring in substrings:
                results = self.app.find_string_ids(substring, st.id)
                if string_id not in results:
                    msg = "Not found: substring '{}' from string '{}'".format(repr(substring), repr(string))
                    print(msg)
                    errors.append(msg)

        # Check for errors.
        self.assertFalse(errors, "\n".join(errors))


def get_all_substrings(s):
    length = len(s)
    return (s[i:j] for i in six.moves.xrange(length) for j in six.moves.xrange(i + 1, length + 1))


worker_app = None


def pool_initializer():
    global worker_app
    worker_app = SuffixTreeApplicationWithCassandra()


def add_string_to_suffix_tree(args):
    # random.seed()
    string, string_id, suffix_tree_id = args
    print("")
    print("Adding string to suffix tree: {}: {}".format(string_id, repr(string[:100])))
    try:
        assert isinstance(worker_app, AbstractSuffixTreeApplication)
        suffix_tree = worker_app.get_suffix_tree(suffix_tree_id)
        assert isinstance(suffix_tree, GeneralizedSuffixTree)
        started = datetime.datetime.now()
        suffix_tree.add_string(string, string_id)
        print(" - added string in: {}".format(datetime.datetime.now() - started))
    except Exception as e:
        msg = traceback.format_exc()
        print(" - failed to add string: {}".format(msg))
        return Exception((e, msg, string, string_id))
    return string_id



from eventsourcing.application.example.with_cassandra import ExampleApplicationWithCassandra
from eventsourcing.infrastructure.stored_event_repos.with_cassandra import create_cassandra_keyspace_and_tables, \
    drop_cassandra_keyspace
from eventsourcing.tests.unit_test_cases_cassandra import CassandraTestCase
from eventsourcing.tests.unit_test_cases_example_application import ExampleApplicationTestCase


class TestApplicationWithCassandra(CassandraTestCase, ExampleApplicationTestCase):

    def create_app(self):
        return ExampleApplicationWithCassandra()

    def setUp(self):
        super(TestApplicationWithCassandra, self).setUp()
        create_cassandra_keyspace_and_tables()

    def tearDown(self):
        drop_cassandra_keyspace()
        super(TestApplicationWithCassandra, self).tearDown()



from __future__ import unicode_literals

import datetime
import importlib
import json
import uuid
from abc import ABCMeta, abstractmethod
from collections import namedtuple

import dateutil.parser
import six

from eventsourcing.domain.model.entity import EventSourcedEntity
from eventsourcing.domain.model.events import topic_from_domain_class, resolve_domain_topic, resolve_attr, DomainEvent

try:
    import numpy
    from six import BytesIO
except:
    numpy = None

StoredEvent = namedtuple('StoredEvent', ['event_id', 'stored_entity_id', 'event_topic', 'event_attrs'])

EntityVersion = namedtuple('EntityVersion', ['entity_version_id', 'event_id'])

class AbstractTranscoder(six.with_metaclass(ABCMeta)):

    @abstractmethod
    def serialize(self, domain_event):
        """Returns a stored event, for the given domain event."""

    @abstractmethod
    def deserialize(self, stored_event):
        """Returns a domain event, for the given stored event."""


class Transcoder(AbstractTranscoder):

    serialize_without_json = False
    serialize_with_uuid1 = True

    def __init__(self, json_encoder_cls=None, json_decoder_cls=None, cipher=None, always_encrypt=False):
        self.json_encoder_cls = json_encoder_cls
        self.json_decoder_cls = json_decoder_cls
        self.cipher = cipher
        self.always_encrypt = always_encrypt

    def serialize(self, domain_event):
        return serialize_domain_event(
            domain_event,
            json_encoder_cls=self.json_encoder_cls,
            without_json=self.serialize_without_json,
            with_uuid1=self.serialize_with_uuid1,
            cipher=self.cipher,
            always_encrypt=self.always_encrypt,
        )

    def deserialize(self, stored_event):
        return deserialize_domain_event(
            stored_event,
            json_decoder_cls=self.json_decoder_cls,
            without_json=self.serialize_without_json,
            with_uuid1=self.serialize_with_uuid1,
            cipher=self.cipher,
            always_encrypt=self.always_encrypt,
        )


def serialize_domain_event(domain_event, json_encoder_cls=None, without_json=False, with_uuid1=False, cipher=None,
                           always_encrypt=False):
    """
    Serializes a domain event into a stored event.
    """
    # assert isinstance(domain_event, DomainEvent)

    # Copy the state of the domain event.
    event_attrs = domain_event.__dict__.copy()

    # Get, or make, the domain event ID.
    if with_uuid1:
        event_id = event_attrs.pop('domain_event_id')
    else:
        event_id = uuid.uuid4().hex

    # Make stored entity ID and topic.
    stored_entity_id = make_stored_entity_id(id_prefix_from_event(domain_event), domain_event.entity_id)
    event_topic = topic_from_domain_class(type(domain_event))

    # Serialise event attributes to JSON, optionally encrypted with cipher.
    if not without_json:

        if json_encoder_cls is None:
            json_encoder_cls = ObjectJSONEncoder

        event_attrs = json.dumps(event_attrs, separators=(',', ':'), sort_keys=True, cls=json_encoder_cls)

        if always_encrypt or domain_event.__class__.always_encrypt:
            if cipher is None:
                raise ValueError("Can't encrypt without a cipher")
            event_attrs = cipher.encrypt(event_attrs)

    # Return a named tuple.
    return StoredEvent(
        event_id=event_id,
        stored_entity_id=stored_entity_id,
        event_topic=event_topic,
        event_attrs=event_attrs,
    )


def deserialize_domain_event(stored_event, json_decoder_cls=None, without_json=False, with_uuid1=False, cipher=None,
                             always_encrypt=False):
    """
    Recreates original domain event from stored event topic and event attrs.
    """
    assert isinstance(stored_event, StoredEvent)

    # Get the domain event class from the topic.
    event_class = resolve_domain_topic(stored_event.event_topic)

    if not isinstance(event_class, type):
        raise ValueError("Event class is not a type: {}".format(event_class))

    if not issubclass(event_class, DomainEvent):
        raise ValueError("Event class is not a DomainEvent: {}".format(event_class))

    # Deserialize event attributes from JSON, optionally decrypted with cipher.
    event_attrs = stored_event.event_attrs
    if not without_json:

        if json_decoder_cls is None:
            json_decoder_cls = ObjectJSONDecoder

        if always_encrypt or event_class.always_encrypt:
            if cipher is None:
                raise ValueError("Can't decrypt stored event without a cipher")
            event_attrs = cipher.decrypt(event_attrs)

        event_attrs = json.loads(event_attrs, cls=json_decoder_cls)

    # Set the domain event ID.
    if with_uuid1:
        event_attrs['domain_event_id'] = stored_event.event_id

    # Reinstantiate and return the domain event object.
    try:
        domain_event = object.__new__(event_class)
        domain_event.__dict__.update(event_attrs)
    except TypeError:
        raise ValueError("Unable to instantiate class '{}' with data '{}'"
                         "".format(stored_event.event_topic, event_attrs))

    return domain_event


class ObjectJSONEncoder(json.JSONEncoder):

    def default(self, obj):
        try:
            return super(ObjectJSONEncoder, self).default(obj)
        except TypeError as e:
            if "not JSON serializable" not in str(e):
                raise
            if isinstance(obj, datetime.datetime):
                return {'ISO8601_datetime': obj.strftime('%Y-%m-%dT%H:%M:%S.%f%z')}
            if isinstance(obj, datetime.date):
                return {'ISO8601_date': obj.isoformat()}
            if numpy is not None and isinstance(obj, numpy.ndarray) and obj.ndim == 1:
                memfile = BytesIO()
                numpy.save(memfile, obj)
                memfile.seek(0)
                serialized = json.dumps(memfile.read().decode('latin-1'))
                d = {
                    '__ndarray__': serialized,
                }
                return d
            else:
                d = {
                    '__class__': obj.__class__.__qualname__,
                    '__module__': obj.__module__,
                }
                return d


class ObjectJSONDecoder(json.JSONDecoder):

    def __init__(self, **kwargs):
        super(ObjectJSONDecoder, self).__init__(object_hook=ObjectJSONDecoder.from_jsonable, **kwargs)

    @staticmethod
    def from_jsonable(d):
        if '__ndarray__' in d:
            return ObjectJSONDecoder._decode_ndarray(d)
        elif '__class__' in d and '__module__' in d:
            return ObjectJSONDecoder._decode_class(d)
        elif 'ISO8601_datetime' in d:
            return ObjectJSONDecoder._decode_datetime(d)
        elif 'ISO8601_date' in d:
            return ObjectJSONDecoder._decode_date(d)
        return d

    @staticmethod
    def _decode_ndarray(d):
        serialized = d['__ndarray__']
        memfile = BytesIO()
        memfile.write(json.loads(serialized).encode('latin-1'))
        memfile.seek(0)
        return numpy.load(memfile)

        # return numpy.array(obj_data, d['dtype']).reshape(d['shape'])

    @staticmethod
    def _decode_class(d):
        class_name = d.pop('__class__')
        module_name = d.pop('__module__')
        module = importlib.import_module(module_name)
        cls = resolve_attr(module, class_name)
        try:
            obj = cls(**d)
        except Exception:
            obj = cls()
            for attr, value in d.items():
                obj.__dict__[attr] = ObjectJSONDecoder.from_jsonable(value)
        return obj

    @staticmethod
    def _decode_date(d):
        return datetime.datetime.strptime(d['ISO8601_date'], '%Y-%m-%d').date()

    @staticmethod
    def _decode_datetime(d):
        return dateutil.parser.parse(d['ISO8601_datetime'])


def deserialize_domain_entity(entity_topic, entity_attrs):
    """
    Return a new domain entity object from a given topic (a string) and attributes (a dict).
    """

    # Get the domain entity class from the entity topic.
    domain_class = resolve_domain_topic(entity_topic)

    # Instantiate the domain entity class.
    entity = object.__new__(domain_class)

    # Set the attributes.
    entity.__dict__.update(entity_attrs)

    # Return a new domain entity object.
    return entity


def make_stored_entity_id(id_prefix, entity_id):
    return '{}::{}'.format(id_prefix, entity_id)


def id_prefix_from_event(domain_event):
    assert isinstance(domain_event, DomainEvent), type(domain_event)
    return id_prefix_from_event_class(type(domain_event))


def id_prefix_from_event_class(domain_event_class):
    assert issubclass(domain_event_class, DomainEvent), type(domain_event_class)
    return domain_event_class.__qualname__.split('.')[0]


def id_prefix_from_entity(domain_entity):
    assert isinstance(domain_entity, EventSourcedEntity)
    return id_prefix_from_entity_class(type(domain_entity))


def id_prefix_from_entity_class(domain_class):
    assert issubclass(domain_class, EventSourcedEntity)
    return domain_class.__name__



import itertools
from collections import defaultdict

from eventsourcing.domain.services.eventstore import AbstractStoredEventRepository
from eventsourcing.domain.services.transcoding import EntityVersion
from eventsourcing.exceptions import ConcurrencyError, EntityVersionDoesNotExist
from eventsourcing.utils.time import timestamp_from_uuid


class PythonObjectsStoredEventRepository(AbstractStoredEventRepository):

    def __init__(self, *args, **kwargs):
        super(PythonObjectsStoredEventRepository, self).__init__(*args, **kwargs)
        self._by_id = {}
        self._by_stored_entity_id = {}
        self._entity_versions = defaultdict(lambda: defaultdict(lambda: itertools.chain([0], itertools.cycle([1]))))
        self._event_id_by_entity_version_id = {}

    def write_version_and_event(self, new_stored_event, new_version_number=None, max_retries=3, artificial_failure_rate=0):
        # Put the event in the various dicts.
        stored_entity_id = new_stored_event.stored_entity_id
        if self.always_write_entity_version and new_version_number is not None:
            versions = self._entity_versions[stored_entity_id]
            if next(versions[new_version_number]) != 0:
                raise ConcurrencyError("New version {} for entity {} already exists"
                                       "".format(new_version_number, stored_entity_id))
            entity_version_id = self.make_entity_version_id(stored_entity_id, new_version_number)
            self._event_id_by_entity_version_id[entity_version_id] = new_stored_event.event_id

        # Remove entity if it's a discarded event.
        if new_stored_event.event_topic.endswith('Discarded'):
            self.remove_entity(stored_entity_id)

        # Otherwise add event to entity's list of events.
        else:
            # Index by entity ID.
            if stored_entity_id not in self._by_stored_entity_id:
                self._by_stored_entity_id[stored_entity_id] = []
            self._by_stored_entity_id[stored_entity_id].append(new_stored_event)

            # Index by event ID.
            self._by_id[new_stored_event.event_id] = new_stored_event

    def get_entity_version(self, stored_entity_id, version_number):
        versions = self._entity_versions[stored_entity_id]
        if version_number not in versions:
            raise EntityVersionDoesNotExist()
        entity_version_id = self.make_entity_version_id(stored_entity_id, version_number)
        return EntityVersion(
            entity_version_id=entity_version_id,
            event_id=self._event_id_by_entity_version_id[entity_version_id]
        )

    def remove_entity(self, stored_entity_id):
        if stored_entity_id in self._by_stored_entity_id:
            for stored_event in self._by_stored_entity_id.pop(stored_entity_id):
                del(self._by_id[stored_event.event_id])
        if self.always_write_entity_version:
            del(self._entity_versions[stored_entity_id])

    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, query_ascending=True,
                          results_ascending=True):
        if stored_entity_id not in self._by_stored_entity_id:
            return []
        else:
            # Get a copy of the list of stored events for this entity.
            stored_events = self._by_stored_entity_id[stored_entity_id][:]

            # Stored event, here, are in ascending order (because they get appended to a list).
            if not query_ascending:
                stored_events.reverse()

            # Get timestamps (floats) from the UUID hex strings (chronologically to compare events).
            if after is None:
                after_timestamp = None
            else:
                after_timestamp = timestamp_from_uuid(after)
            if until is None:
                until_timestamp = None
            else:
                until_timestamp = timestamp_from_uuid(until)

            # Start counting events (needed to stop when limit is reached).
            count = 0

            # Initialise the query results.
            query_results = []

            # Iterate over the stored events, excluding things that don't match.
            for event in stored_events:
                event_timestamp = timestamp_from_uuid(event.event_id)

                if limit is not None and count >= limit:
                    break

                # Exclude if earlier than the 'after' time.
                if after_timestamp:
                    if query_ascending:
                        if event_timestamp <= after_timestamp:
                            continue
                    else:
                        if event_timestamp < after_timestamp:
                            continue

                # Exclude if later than the 'until' time.
                if until_timestamp:
                    if query_ascending:
                        if event_timestamp > until_timestamp:
                            continue
                    else:
                        if event_timestamp >= until_timestamp:
                            continue

                query_results.append(event)
                count += 1

                if results_ascending != query_ascending:
                    query_results.reverse()

            return query_results



from eventsourcing.application.base import EventSourcingApplication
from eventsourcing.infrastructure.stored_event_repos.with_python_objects import PythonObjectsStoredEventRepository


class EventSourcingWithPythonObjects(EventSourcingApplication):

    def create_stored_event_repo(self, *args, **kwargs):
        return PythonObjectsStoredEventRepository(*args, **kwargs)



from eventsourcing.tests.unit_test_cases import BasicStoredEventRepositoryTestCase, SimpleStoredEventIteratorTestCase, \
    ThreadedStoredEventIteratorTestCase
from eventsourcing.tests.unit_test_cases_python_objects import PythonObjectsTestCase


class TestPythonObjectsStoredEventRepository(PythonObjectsTestCase, BasicStoredEventRepositoryTestCase):
    pass


class TestSimpleStoredEventIteratorWithPythonObjects(PythonObjectsTestCase, SimpleStoredEventIteratorTestCase):
    pass


class TestThreadedStoredEventIteratorWithPythonObjects(PythonObjectsTestCase, ThreadedStoredEventIteratorTestCase):
    pass


# Todo: Revisit this, but with threading rather than multiprocessing because data is stored in process.
# class TestConcurrentStoredEventRepositoryWithPythonObjects(PythonObjectsTestCase, ConcurrentStoredEventRepositoryTestCase):
#     pass



from eventsourcing.application.example.base import ExampleApplication
from eventsourcing.application.example.with_cassandra import ExampleApplicationWithCassandra
from eventsourcing.application.example.with_pythonobjects import ExampleApplicationWithPythonObjects
from eventsourcing.domain.model.example import Example
from eventsourcing.exceptions import ConcurrencyError
from eventsourcing.tests.unit_test_cases_cassandra import CassandraTestCase
from eventsourcing.tests.unit_test_cases_example_application import ExampleApplicationTestCase


class TestFastForward(CassandraTestCase, ExampleApplicationTestCase):
    def create_app(self):
        return ExampleApplicationWithCassandra()
        # return ExampleApplicationWithPythonObjects()

    def test(self):
        assert isinstance(self.app, ExampleApplication)
        example = self.app.register_new_example(1, 2)
        instance1 = self.app.example_repo[example.id]
        instance2 = self.app.example_repo[example.id]

        assert isinstance(instance1, Example)
        assert isinstance(instance2, Example)

        self.assertEqual(instance1.version, 1)
        self.assertEqual(instance2.version, 1)

        # Evolve instance1 by two versions.
        instance1.beat_heart()
        # instance1.beat_heart()
        self.assertEqual(instance1.version, 2)

        # Try to evolve instance2 from the same version.
        # - check it raises a concurrency error
        preop_state = instance2.__dict__.copy()
        with self.assertRaises(ConcurrencyError):
            instance2.beat_heart()

        # Reset instance2 to its pre-op state.
        instance2.__dict__.update(preop_state)
        self.assertEqual(instance2.version, 1)

        # Fast forward instance2 from pre-op state.
        instance3 = self.app.example_repo.fastforward(instance2)
        self.assertEqual(instance2.version, 1)
        self.assertEqual(instance3.version, 2)

        # Try again to beat heart.
        instance3.beat_heart()
        self.assertEqual(instance3.version, 3)

        # Try to evolve instance1 from its stale version.
        preop_state = instance1.__dict__.copy()
        with self.assertRaises(ConcurrencyError):
            instance1.beat_heart()

        # Reset instance1 to pre-op state.
        instance1.__dict__.update(preop_state)

        # Fast forward instance1 from pre-op state.
        instance4 = self.app.example_repo.fastforward(instance1)
        self.assertEqual(instance1.version, 2)
        self.assertEqual(instance4.version, 3)

        # Try again to beat heart.
        instance4.beat_heart()
        self.assertEqual(instance4.version, 4)



# coding=utf-8
from __future__ import unicode_literals

import os
import unittest

from eventsourcing.contrib.suffixtrees.domain.model.suffixtree import register_new_suffix_tree, SuffixTree,\
    SuffixTreeApplication
from eventsourcing.tests.unit_test_fixtures_suffix_tree_text import LONG_TEXT

LONG_TEXT_FIXTURE_PATH = os.path.join(os.path.dirname(__file__), 'test_suffix_tree.txt')


class SuffixTreeTest(unittest.TestCase):

    def setUp(self):
        self.app = SuffixTreeApplication()

    def tearDown(self):
        self.app.close()

    def test_empty_string(self):
        st = register_new_suffix_tree()
        assert isinstance(st, SuffixTree)
        st.add_string('')
        self.assertEqual(self.app.find_substring('not there', st.id), -1)
        self.assertEqual(self.app.find_substring('', st.id), -1)
        self.assertFalse(self.app.has_substring('not there', st.id))
        self.assertFalse(self.app.has_substring('', st.id))

    def test_repeated_string(self):
        st = register_new_suffix_tree()
        st.add_string("aaa")
        self.assertEqual(self.app.find_substring('a', st.id), 0)
        self.assertEqual(self.app.find_substring('aa', st.id), 0)
        self.assertEqual(self.app.find_substring('aaa', st.id), 0)
        self.assertEqual(self.app.find_substring('b', st.id), -1)
        self.assertTrue(self.app.has_substring('a', st.id))
        self.assertTrue(self.app.has_substring('aa', st.id))
        self.assertTrue(self.app.has_substring('aaa', st.id))
        self.assertFalse(self.app.has_substring('aaaa', st.id))
        self.assertFalse(self.app.has_substring('b', st.id))
        # case sensitive by default
        self.assertFalse(self.app.has_substring('A', st.id))

    def test_mississippi(self):
        st = register_new_suffix_tree()
        st.add_string("mississippi")
        self.assertEqual(self.app.find_substring('a', st.id), -1)
        self.assertEqual(self.app.find_substring('m', st.id), 0)
        self.assertEqual(self.app.find_substring('i', st.id), 1)

    def test_long_string(self):
        st = register_new_suffix_tree()
        st.add_string(LONG_TEXT)
        self.assertEqual(self.app.find_substring('Ukkonen', st.id), 1498)
        self.assertEqual(self.app.find_substring('Optimal', st.id), 11074)
        self.assertFalse(self.app.has_substring('ukkonen', st.id))

    def test_case_insensitivity(self):
        st = register_new_suffix_tree(case_insensitive=True)
        st.add_string(LONG_TEXT)
        self.assertEqual(self.app.find_substring('ukkonen', st.id), 1498)
        self.assertEqual(self.app.find_substring('Optimal', st.id), 1830)



from eventsourcing.tests.unit_test_cases import BasicStoredEventRepositoryTestCase, \
    ConcurrentStoredEventRepositoryTestCase, SimpleStoredEventIteratorTestCase, ThreadedStoredEventIteratorTestCase
from eventsourcing.tests.unit_test_cases_cassandra import CassandraStoredEventRepoTestCase


class TestCassandraStoredEventRepository(CassandraStoredEventRepoTestCase, BasicStoredEventRepositoryTestCase):
    pass


class TestSimpleStoredEventIteratorWithCassandra(CassandraStoredEventRepoTestCase, SimpleStoredEventIteratorTestCase):
    pass


class TestThreadedStoredEventIteratorWithCassandra(CassandraStoredEventRepoTestCase, ThreadedStoredEventIteratorTestCase):
    pass


class TestConcurrentStoredEventRepositoryWithCassandra(CassandraStoredEventRepoTestCase, ConcurrentStoredEventRepositoryTestCase):
    pass



import os
from random import random
from time import sleep

import cassandra.cqlengine.connection
import six
from cassandra import ConsistencyLevel, AlreadyExists, DriverException, OperationTimedOut, InvalidRequest
from cassandra.auth import PlainTextAuthProvider
from cassandra.cqlengine.management import sync_table, create_keyspace_simple, drop_keyspace
from cassandra.cqlengine.models import Model, columns
from cassandra.cqlengine.query import LWTException

from eventsourcing.domain.services.eventstore import AbstractStoredEventRepository
from eventsourcing.domain.services.transcoding import StoredEvent, EntityVersion
from eventsourcing.exceptions import ConcurrencyError, EntityVersionDoesNotExist
from eventsourcing.infrastructure.stored_event_repos.threaded_iterator import ThreadedStoredEventIterator

DEFAULT_CASSANDRA_KEYSPACE = os.getenv('CASSANDRA_KEYSPACE', 'eventsourcing')
DEFAULT_CASSANDRA_CONSISTENCY_LEVEL = os.getenv('CASSANDRA_CONSISTENCY_LEVEL', 'LOCAL_QUORUM')
DEFAULT_CASSANDRA_HOSTS = [h.strip() for h in os.getenv('CASSANDRA_HOSTS', 'localhost').split(',')]
DEFAULT_CASSANDRA_PORT = int(os.getenv('CASSANDRA_PORT', 9042))
DEFAULT_CASSANDRA_PROTOCOL_VERSION = int(os.getenv('CASSANDRA_PROTOCOL_VERSION', 3))


class CqlEntityVersion(Model):
    """Stores the stored entity ID and entity version as a partition key."""

    __table_name__ = 'entity_versions'

    # Makes sure we can't write the same version twice,
    # helps to implement optimistic concurrency control.
    _if_not_exists = True

    # Entity-version identifier (a string).
    r = columns.Text(partition_key=True)

    # Stored event ID (normally a uuid1().hex).
    v = columns.Text()

    # Because models with one column break Cassandra driver 3.5.0.
    x = columns.Text(default='')


class CqlStoredEvent(Model):

    __table_name__ = 'stored_events'

    # Stored entity ID (normally a string, with the entity type name at the start)
    n = columns.Text(partition_key=True)

    # Stored event ID (normally a UUID1)
    v = columns.TimeUUID(clustering_order='DESC', primary_key=True)

    # Stored event topic (path to the event object class)
    t = columns.Text(required=True)

    # Stored event attributes (the entity object __dict__)
    a = columns.Text(required=True)


def to_cql(stored_event):
    assert isinstance(stored_event, StoredEvent), stored_event
    return CqlStoredEvent(
        n=stored_event.stored_entity_id,
        v=stored_event.event_id,
        t=stored_event.event_topic,
        a=stored_event.event_attrs
    )


def from_cql(cql_stored_event):
    assert isinstance(cql_stored_event, CqlStoredEvent), cql_stored_event
    return StoredEvent(
        stored_entity_id=cql_stored_event.n,
        event_id=cql_stored_event.v.hex,
        event_topic=cql_stored_event.t,
        event_attrs=cql_stored_event.a
    )


class CassandraStoredEventRepository(AbstractStoredEventRepository):

    @property
    def iterator_class(self):
        return ThreadedStoredEventIterator

    def write_version_and_event(self, new_stored_event, new_version_number=None, max_retries=3, artificial_failure_rate=0):
        """
        Writes entity version if not exists, and then writes the stored event.
        """
        # Write the next version.
        stored_entity_id = new_stored_event.stored_entity_id
        new_entity_version = None
        if self.always_write_entity_version and new_version_number is not None:
            assert isinstance(new_version_number, six.integer_types)
            #  - uses the "if not exists" optimistic concurrency control feature
            #    of Cassandra, hence this operation is assumed to succeed only once
            new_entity_version_id = self.make_entity_version_id(stored_entity_id, new_version_number)
            new_entity_version = CqlEntityVersion(
                r=new_entity_version_id,
                v=new_stored_event.event_id,
            )
            try:
                new_entity_version.save()
            except LWTException as e:
                raise ConcurrencyError("Version {} of entity {} already exists: {}".format(new_entity_version, stored_entity_id, e))

        # Increased latency here causes increased contention.
        #  - used for testing concurrency exceptions
        if artificial_failure_rate:
            sleep(artificial_failure_rate)

        # Write the stored event into the database.
        try:
            retries = max_retries
            while True:
                try:
                    # Instantiate a Cassandra CQL engine object.
                    cql_stored_event = to_cql(new_stored_event)

                    # Optionally mimic an unreliable save() operation.
                    #  - used for testing retries
                    if artificial_failure_rate and (random() > 1 - artificial_failure_rate):
                        raise DriverException("Artificial failure")

                    # Save the event.
                    cql_stored_event.save()

                except DriverException:

                    if retries <= 0:
                        # Raise the error after retries exhausted.
                        raise
                    else:
                        # Otherwise retry.
                        retries -= 1
                        sleep(0.05 + 0.1 * random())
                else:
                    break

        except DriverException as event_write_error:
            # If we get here, we're in trouble because the version has been
            # written, but perhaps not the event, so the entity may be broken
            # because it might not be possible to get the entity with version
            # number high enough to pass the optimistic concurrency check
            # when storing subsequent events.

            # Back off for a little bit.
            sleep(0.1)
            try:
                # If the event actually exists, despite the exception, all is well.
                CqlStoredEvent.get(n=new_stored_event.stored_entity_id, v=new_stored_event.event_id)

            except CqlStoredEvent.DoesNotExist:
                # Otherwise, try harder to recover by removing the new version.
                if new_entity_version is not None:
                    retries = max_retries * 3
                    while True:
                        try:
                            # Optionally mimic an unreliable delete() operation.
                            #  - used for testing retries
                            if artificial_failure_rate and (random() > 1 - artificial_failure_rate):
                                raise DriverException("Artificial failure")

                            # Delete the new entity version.
                            new_entity_version.delete()

                        except DriverException as version_delete_error:
                            # It's not going very well, so maybe retry.
                            if retries <= 0:
                                # Raise when retries are exhausted.
                                raise Exception("Unable to delete version {} of entity {} after failing to write"
                                                "event: event write error {}: version delete error {}"
                                                .format(new_entity_version, stored_entity_id,
                                                        event_write_error, version_delete_error))
                            else:
                                # Otherwise retry.
                                retries -= 1
                                sleep(0.05 + 0.1 * random())
                        else:
                            # The entity version was deleted, all is well.
                            break
                raise event_write_error

    def get_entity_version(self, stored_entity_id, version_number):
        entity_version_id = self.make_entity_version_id(stored_entity_id, version_number)
        try:
            cql_entity_version = CqlEntityVersion.get(r=entity_version_id)
        except CqlEntityVersion.DoesNotExist:
            raise EntityVersionDoesNotExist()
        assert isinstance(cql_entity_version, CqlEntityVersion)
        return EntityVersion(
            entity_version_id=entity_version_id,
            event_id=cql_entity_version.v,
        )

    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, query_ascending=True,
                          results_ascending=True):
        query = CqlStoredEvent.objects.filter(n=stored_entity_id)

        if query_ascending:
            query = query.order_by('v')

        if after is not None:
            if query_ascending:
                query = query.filter(v__gt=after)
            else:
                query = query.filter(v__gte=after)
        if until is not None:
            if query_ascending:
                query = query.filter(v__lte=until)
            else:
                query = query.filter(v__lt=until)

        if limit is not None:
            query = query.limit(limit)

        events = self.map(from_cql, query)
        events = list(events)

        if results_ascending != query_ascending:
            events.reverse()

        return events


def get_cassandra_setup_params(hosts=DEFAULT_CASSANDRA_HOSTS, consistency=DEFAULT_CASSANDRA_CONSISTENCY_LEVEL,
                               default_keyspace=DEFAULT_CASSANDRA_KEYSPACE, port=DEFAULT_CASSANDRA_PORT,
                               protocol_version=DEFAULT_CASSANDRA_PROTOCOL_VERSION, username=None, password=None):

    # Construct an "auth provider" object.
    if username and password:
        auth_provider = PlainTextAuthProvider(username, password)
    else:
        auth_provider = None

    # Resolve the consistency level to an object, if it's a string.
    if isinstance(consistency, six.string_types):
        try:
            consistency = getattr(ConsistencyLevel, consistency.upper())
        except AttributeError:
            msg = "Cassandra consistency level '{}' not found.".format(consistency)
            raise Exception(msg)

    return auth_provider, hosts, consistency, default_keyspace, port, protocol_version


def setup_cassandra_connection(auth_provider, hosts, consistency, default_keyspace, port, protocol_version):
    cassandra.cqlengine.connection.setup(
        hosts=hosts,
        consistency=consistency,
        default_keyspace=default_keyspace,
        port=port,
        auth_provider=auth_provider,
        # protocol_version=protocol_version,
        lazy_connect=True,
        retry_connect=True,
    )


def create_cassandra_keyspace_and_tables(keyspace=DEFAULT_CASSANDRA_KEYSPACE, replication_factor=1):
    # Avoid warnings about this variable not being set.
    os.environ['CQLENG_ALLOW_SCHEMA_MANAGEMENT'] = '1'
    try:
        create_keyspace_simple(keyspace, replication_factor=replication_factor)
    except AlreadyExists:
        pass
    else:
        sync_table(CqlStoredEvent)
        sync_table(CqlEntityVersion)


def drop_cassandra_keyspace(keyspace=DEFAULT_CASSANDRA_KEYSPACE):
    max_retries = 3
    tried = 0
    while True:
        try:
            drop_keyspace(keyspace)
            break
        except InvalidRequest:
            break
        except OperationTimedOut:
            tried += 1
            if tried <= max_retries:
                sleep(0.5)
                continue
            else:
                raise


def shutdown_cassandra_connection():
    if cassandra.cqlengine.connection.session:
        cassandra.cqlengine.connection.session.shutdown()
    if cassandra.cqlengine.connection.cluster:
        cassandra.cqlengine.connection.cluster.shutdown()



from eventsourcing.application.example.with_cassandra import ExampleApplicationWithCassandra
from eventsourcing.domain.services.cipher import AESCipher
from eventsourcing.tests.test_application_with_cassandra import TestApplicationWithCassandra


class TestApplicationWithEncryption(TestApplicationWithCassandra):

    def create_app(self):
        cipher = AESCipher(aes_key='0123456789abcdef')
        return ExampleApplicationWithCassandra(cipher=cipher, always_encrypt_stored_events=True)



from eventsourcing.infrastructure.stored_event_repos.with_cassandra import CassandraStoredEventRepository, \
    setup_cassandra_connection, get_cassandra_setup_params, create_cassandra_keyspace_and_tables, \
    drop_cassandra_keyspace
from eventsourcing.tests.unit_test_cases import AbstractTestCase


class CassandraTestCase(AbstractTestCase):

    def setUp(self):
        super(CassandraTestCase, self).setUp()
        create_cassandra_keyspace_and_tables()

    def tearDown(self):
        drop_cassandra_keyspace()
        # shutdown_cassandra_connection()
        super(CassandraTestCase, self).tearDown()


class CassandraStoredEventRepoTestCase(CassandraTestCase):

    def setUp(self):
        setup_cassandra_connection(*get_cassandra_setup_params())
        super(CassandraStoredEventRepoTestCase, self).setUp()

    @property
    def stored_event_repo(self):
        try:
            return self._stored_event_repo
        except AttributeError:
            stored_event_repo = CassandraStoredEventRepository()
            self._stored_event_repo = stored_event_repo
            return stored_event_repo
